/*
 *
 *  ao_oss.c
 *
 *      Original Copyright (C) Aaron Holtzman - May 1999
 *      Modifications Copyright (C) Stan Seibert - July 2000, June 2001
 *
 *  This file is part of libao, a cross-platform library.  See
 *  README for a history of this source code.
 *
 *  libao is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *
 *  libao is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with GNU Make; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
   --
    Changes:
    - added "vol" option to optionally set the volume in the driver
    - changed to nonblocking writes
    - added call to SETFRAGMENT ioctl
    by TiVo Inc.
   --
 */

#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <math.h>
#if defined(__OpenBSD__) || defined(__NetBSD__)
#include <soundcard.h>
#elif defined(__FreeBSD__)
#include <machine/soundcard.h>
#else
#include <sys/soundcard.h>
#endif
#include <sys/ioctl.h>

#include "ao/ao.h"

#ifndef OSS_DEVICE1
# define OSS_DEVICE1 "/dev/sound/dsp"
#endif
#ifndef OSS_DEVICE2
# define OSS_DEVICE2 "/dev/dsp"
#endif


static char *ao_oss_options[] = {"dsp","vol"};
static ao_info ao_oss_info =
{
	AO_TYPE_LIVE,
	"OSS audio driver output ",
	"oss",
	"Aaron Holtzman <aholtzma@ess.engr.uvic.ca>",
	"Outputs audio to the Open Sound System driver.",
	AO_FMT_NATIVE,
	20,
	ao_oss_options,
	2
};


typedef struct ao_oss_internal {
	char *dev;
	int fd;
        int vol;
} ao_oss_internal;


/*
 * open either the devfs device or the traditional device and return a
 * file handle.  Also strdup() path to the selected device into
 * *dev_path.  Assumes that *dev_path does not need to be free()'ed
 * initially.
 *
 * If BROKEN_OSS is defined, this opens the device in non-blocking
 * mode at first in order to prevent deadlock caused by ALSA's OSS
 * emulation and some OSS drivers if the device is already in use.  If
 * blocking is non-zero, we remove the blocking flag if possible so
 * that the device can be used for actual output.
 */
int _open_default_oss_device (char **dev_path, int blocking)
{
	int fd;
	char *err = NULL;
	char *dev = NULL;

	/* default: first try the devfs path */
	*dev_path = strdup(OSS_DEVICE1);
#ifdef BROKEN_OSS
	fd = open(*dev_path, O_WRONLY | O_NONBLOCK);
#else
	fd = open(*dev_path, O_WRONLY);
#endif /* BROKEN_OSS */

	/* then try the original dsp path */
	if(fd < 0) 
	{
		/* no? then try the traditional path */
		err = strdup(strerror(errno));
		dev = strdup(*dev_path);
		free(*dev_path);
		*dev_path = strdup(OSS_DEVICE2);
#ifdef BROKEN_OSS
		fd = open(*dev_path, O_WRONLY | O_NONBLOCK);
#else
		fd = open(*dev_path, O_WRONLY);
#endif /* BROKEN_OSS */
	}

#ifdef BROKEN_OSS
	/* Now have to remove the O_NONBLOCK flag if so instructed. */
	if (fd > 0 && blocking) {
		if (fcntl(fd, F_SETFL, 0) < 0) { /* Remove O_NONBLOCK */
			/* If we can't go to blocking mode, we can't use
			   this device */
			close(fd);
			fd = -1;
		}
	}
#endif /* BROKEN_OSS */

	/* Deal with error cases */
	if(fd < 0) 
	{
	  /*			fprintf(stderr,
				"libao - error: Could not open either default device:\n"
				"  %s - %s\n"
				"  %s - %s\n",
				err, dev,
				strerror(errno), *dev_path); */
		free(err);
		free(dev);
		free(*dev_path);
		*dev_path = NULL;
	}

	return fd;
}


int ao_oss_test()
{
	char *dev_path;
	int fd;

	/* OSS emulation in ALSA will by default cause the open() call
	   to block if the dsp is in use.  This will freeze the default
	   driver detection unless the O_NONBLOCK flag is passed to 
	   open().  We cannot use this flag when we actually open the
	   device for writing because then we will overflow the buffer. */
	if ( (fd = _open_default_oss_device(&dev_path, 0)) < 0 )
		return 0; /* Cannot use this plugin with default parameters */
	else {
		free(dev_path);
		close(fd);
		return 1; /* This plugin works in default mode */
	}
}


ao_info *ao_oss_driver_info(void)
{
	return &ao_oss_info;
}


int ao_oss_device_init(ao_device *device)
{
	ao_oss_internal *internal;

	internal = (ao_oss_internal *) malloc(sizeof(ao_oss_internal));

	if (internal == NULL)	
		return 0; /* Could not initialize device memory */
	
	internal->dev = NULL;
        internal->vol = -1;
	
	device->internal = internal;

	return 1; /* Memory alloc successful */
}

int ao_oss_set_option(ao_device *device, const char *key, const char *value)
{
	ao_oss_internal *internal = (ao_oss_internal *) device->internal;


	if (!strcmp(key, "dsp")) {
		/* Free old string in case "dsp" set twice in options */
		free(internal->dev); 
		internal->dev = strdup(value);
	}

        if (!strcmp(key, "vol")) {
            /* set the volume */
            internal->vol = atoi(value);
        }
                
	return 1;
}


int ao_oss_config( ao_device *device, ao_sample_format *format)
{
    int tmp;
    ao_oss_internal *internal = (ao_oss_internal *) device->internal;
    switch (format->channels)
    {
	case 1: tmp = 0;
            break;
	case 2: tmp = 1;
            break;
	default:
            fprintf(stderr,"libao - Unsupported number of channels: %d.",
                    format->channels);
            return -1;
            break;
    }
	
    if (ioctl(internal->fd,SNDCTL_DSP_STEREO,&tmp) < 0) {
        fprintf(stderr, "libao - OSS cannot set channels to %d\n", 
                format->channels);
        return -1;
    }

    /* To eliminate the need for a swap buffer, we set the device
	   to use whatever byte format the client selected. */
    switch (format->bits)
    {
	case 8: tmp = AFMT_S8;
            break;
        case 16: tmp = device->client_byte_format == AO_FMT_BIG ? 
                     AFMT_S16_BE : AFMT_S16_LE;
        device->driver_byte_format = device->client_byte_format;
        break;
	default:fprintf(stderr,"libao - Unsupported number of bits: %d.",
			format->bits);
        return -1;
    }

    if (ioctl(internal->fd,SNDCTL_DSP_SAMPLESIZE,&tmp) < 0) {
        fprintf(stderr, "libao - OSS cannot set sample size to %d\n",
                format->bits);
        return -1;
    }

    tmp = format->rate;
    /* Some cards aren't too accurate with their clocks and set to the
       exact data rate, but something close.  Fail only if completely out
       of whack. */
/*          fprintf(stderr, "libao - OSS setting rate to %d\n", 
            tmp);*/
    if (ioctl(internal->fd,SNDCTL_DSP_SPEED, &tmp) < 0
        || tmp > 1.01 * format->rate || tmp < 0.99 * format->rate) {
        fprintf(stderr, "libao - OSS cannot set rate to %d\n", 
                format->rate);
        return -1;
    }

    if ( internal->vol >= 0 && internal->vol <= 100)
    {
        uint vol = (uint)internal->vol;
/*            fprintf(stderr, "libao - setting volume to %d\n", vol); */
        uint v=(vol<<8|vol);
        if (ioctl(internal->fd,SOUND_MIXER_WRITE_PCM,&v) < 0 )
        {
            fprintf(stderr, "libao - OSS cannot set volume to %d\n", vol );
            return -1;
        }
    }

    /* TIVO */
    /* Set PCM output buffer size */
    tmp = 14; /* 16k */
    if (ioctl(internal->fd, SNDCTL_DSP_SETFRAGMENT, &tmp) < 0) {
        fprintf(stderr, "libao - OSS cannot set FRAGMENT size\n" );
        return -1;
    }
    else {
//         fprintf(stderr, "libao: SNDCTL_DSP_SETFRAGMENT to %d\n",
//                 ( 1 << ( tmp & 0xffff )));
    }
    /* TIVO */
    
    return 1; /* Open successful */
}

/*
 * open the audio device for writing to
 */
int ao_oss_open(ao_device *device, ao_sample_format *format)
{
	ao_oss_internal *internal = (ao_oss_internal *) device->internal;

	/* Open the device driver */
	
	if (internal->dev != NULL) {
		/* open the user-specified path */
		internal->fd = open(internal->dev, O_WRONLY);

		if(internal->fd < 0) {
                    fprintf(stderr,"libao - %s: Opening audio device %s\n",
                            strerror(errno), internal->dev); 
                    return 0;  /* Cannot open device */
                }
	} else {
		internal->fd = _open_default_oss_device(&internal->dev, 1);
		if (internal->fd < 0)
			return 0;  /* Cannot open default device */
	}

	/* Now set all of the parameters */

        if (ao_oss_config(device, format) < 0 ) {
            close(internal->fd);
            return 0; /* Failed to open device */
        }
        else return 1;
}

/*
 * play the sample to the already opened file descriptor
 */
int ao_oss_play(ao_device *device, const char *output_samples, 
		uint_32 num_bytes)
{
	ao_oss_internal *internal = (ao_oss_internal *) device->internal;

//          audio_buf_info  info;
//          int ios;
//          if ((ios = ioctl(internal->fd,SNDCTL_DSP_GETOSPACE,&info)) >= 0 
//              && info.bytes < 512 )
//          {
//              fprintf(stderr, 
//                      "GETOSPACE: bytes: %d\tfrags: %d\tfragsize: %d\n",
//                      info.bytes, info.fragments, info.fragsize);
//          }
        
        int num_written = 0;
        int fTryBlocking = 0;
        int retval = 0;
        fcntl(internal->fd, F_SETFL, O_NONBLOCK);
	if ( (num_written = write(internal->fd, output_samples, num_bytes) ) < 0) {
            
            if ( errno == EAGAIN ) {
                fTryBlocking = 1;
                num_written = 0;
            }
            
            retval = 0;
        }
	else {
            
            if ( (uint_32)num_written == num_bytes ) {
//                 fprintf(stderr,
//                         "ao_oss_play: num_written = %d, num_bytes = %d (could be underflow?)\n",
//                         num_written, num_bytes);
            }
            else {
                fTryBlocking = 1;
            }
            retval = 1;
        }
        
        if ( fTryBlocking ) {
            // Switch to blocking
            fcntl(internal->fd, F_SETFL, 0);
            write(internal->fd, output_samples + num_written, num_bytes - num_written);
        }
        
        return retval;
}


int ao_oss_close(ao_device *device)
{
	ao_oss_internal *internal = (ao_oss_internal *) device->internal;
	close(internal->fd);

	return 1;
}


void ao_oss_device_clear(ao_device *device)
{
	ao_oss_internal *internal = (ao_oss_internal *) device->internal;

	free(internal->dev);
	free(internal);
}

int ao_oss_device_stop(ao_device *device)
{
	ao_oss_internal *internal = (ao_oss_internal *) device->internal;

        if (ioctl(internal->fd, SNDCTL_DSP_RESET, 0) == -1) {
            perror("ioctl(SNDCTL_DSP_RESET)");
            return 0;
        }

        return 1;
}

ao_functions ao_oss = {
	ao_oss_test,
	ao_oss_driver_info,
	ao_oss_device_init,
	ao_oss_set_option,
	ao_oss_open,
	ao_oss_play,
	ao_oss_close,
	ao_oss_device_clear,
        NULL,
        ao_oss_device_stop
};
