/*
    mpg321 - a fully free clone of mpg123.
    Copyright (C) 2001 Joe Drew
    
    Originally based heavily upon:
    plaympeg - Sample MPEG player using the SMPEG library
    Copyright (C) 1999 Loki Entertainment Software
    
    Also uses some code from
    mad - MPEG audio decoder
    Copyright (C) 2000-2001 Robert Leslie
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   --
    Changes:
    - not stop if it cannot find a file
    - new "@E" output if error
    - add '\t' between ID3 tags
    - change when @I is sent, and what it does with URLs
    - added "SOFTWARE_VOLUME" flag to enable/disable software volume adjustment
      (vs. using the driver)
    - Consolidated remote output through new function mpg321_msg 
      for easier debugging
    - handle errors more gracefully than exit()ing
    - Open a local file using open() instead of MMAP if HAVE_MMAP is not set
    - Added -i to ignore CRC errors
    - Added @P -1 for indicating when file is closed
    - Cleaned up some potential buffer overrun exploits
    - Fixed mmap/munmap so that we ensure the same page-aligned length is
      passed into both
    - added mpg321_printf() and mpg321_log()
      
    by TiVo Inc.
   --
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <libgen.h>
#include <signal.h>
#include <limits.h>
#include <sys/time.h>
#include <assert.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdarg.h>
#include <sys/mman.h>
#include <syslog.h>

#include "getopt.h" /* GNU getopt is needed, so I included it */
#include "mpg321.h"

#include <id3tag.h>

int shuffle_play;
int stop_playing_file = 0;
int quit_now = 0;
char *playlist_file;
ao_device *playdevice=NULL;
mad_timer_t current_time;
mpg321_options options = {
    0,          /* opt */
    NULL,       /* devicetype */
    NULL,       /* device */
    0,          /* seek */
    0,          /* maxframes */
    0,          /* volume */
    0           /* skip_printing_frames */
};
int status = MPG321_STOPPED;
int file_change = 0;

long num_decoder_errors = 0;

/* Forward decls */
char *id3_get_tag (struct id3_tag const *tag, char const *what, unsigned int maxlen);
static int show_id3(struct id3_tag const *tag);

void mpg123_boilerplate()
{
    fprintf(stderr,"High Performance MPEG 1.0/2.0/2.5 Audio Player for Layer 1, 2, and 3.\n"
        "Version " FAKEVERSION " (" VERSIONDATE "). Written and copyrights by Joe Drew.\n"
        "Uses code from various people. See 'README' for more!\n"
        "THIS SOFTWARE COMES WITH ABSOLUTELY NO WARRANTY! USE AT YOUR OWN RISK!\n");
}

/* write out messages to remote clients (if needed)*/
void mpg321_va_msg(char *fmt, va_list ap)
{
#ifdef DEBUG_REMOTE_OUTPUT
    // NOTE: This may not work on all compilers!!!:
    va_list ap2 = ap;
#endif    
    if (options.opt & MPG321_REMOTE_PLAY) {
        vprintf(fmt, ap);
        fflush(stdout);
#ifdef DEBUG_REMOTE_OUTPUT
        {
            static FILE *fp = NULL;
            if ( NULL == fp )
            {
                fp = fopen("/tmp/mpg321_remote.out", "w");
            }
            vfprintf(fp, fmt, ap2);
            fflush(fp);
        }
#endif        
    }
}

void mpg321_msg(char *fmt, ...)
{
    va_list ap;
    if (options.opt & MPG321_REMOTE_PLAY) {
        va_start(ap, fmt);
        mpg321_va_msg(fmt, ap);
        va_end(ap);
    }
}

void mpg321_log(char *fmt, ...)
{
    va_list ap;

    openlog("mpg321", LOG_NDELAY | LOG_PID, LOG_LOCAL0);
    va_start(ap, fmt);
    vsyslog( LOG_LOCAL0 | LOG_INFO, fmt, ap );
    va_end(ap);
    closelog();
}

/* print the error in errno */
void mpg321_error(char *fmt, ...)
{
    va_list ap;
    
    char *err = strerror(errno);
    if (options.opt & MPG321_REMOTE_PLAY) {
        mpg321_msg("@E mpg321 error:%#x: ", errno);
        va_start(ap, fmt);
        mpg321_va_msg(fmt, ap);
        va_end(ap);
        mpg321_msg(" [%s]\n", err);
    }
    else
    {
        fprintf(stderr, "mpg321 error: ");
        va_list ap2;
        va_start(ap2, fmt);
        vfprintf(stderr, fmt, ap2);
        va_end(ap2);
        fprintf(stderr, " [%s]\n", err);
    }
}

/*
 * NAME:	get_id3()
 * DESCRIPTION:	read and parse an ID3 tag from a stream
 */
static
struct id3_tag *get_id3(struct mad_stream *stream, id3_length_t tagsize,
			buffer *playbuf)
{
  struct id3_tag *tag = 0;
  id3_length_t count;
  id3_byte_t const *data;
  id3_byte_t *allocated = 0;
  extern ssize_t fd_blocking_read(int, void*, size_t);
  
  count = stream->bufend - stream->this_frame;

  if (tagsize <= count) {
    data = stream->this_frame;
    mad_stream_skip(stream, tagsize);
  }
  else {
    allocated = malloc(tagsize);
    if (allocated == 0) {
        mpg321_error("get_id3: not enough memory to allocate tag data buffer");
        goto fail;
    }

    memcpy(allocated, stream->this_frame, count);

    while (count < tagsize) {
      int len;

      len = fd_blocking_read(playbuf->fd, allocated + count, tagsize - count);

      if (len == -1) {
          mpg321_error( "stream id3 read" );
          goto fail;
      }

      if (len == 0) {
          //XXX???Sometimes we get a spurious error here when we interrupt
          //XXX   playback to start a new song.  Suppressing until we can 
          //XXX   figure out why 
          // mpg321_error( "EOF while reading tag data" );
          goto fail;
      }

      count += len;
    }

    data = allocated;

    mad_stream_skip(stream, tagsize);
  }

  if ((id3_length_t)playbuf->length >= tagsize)
      playbuf->length -= tagsize;
  
  tag = id3_tag_parse(data, tagsize);

 fail:
  if (allocated)
    free(allocated);

  return tag;
}


/* handle a decode error */
enum mad_flow decode_error(void *data, struct mad_stream *stream,
			   struct mad_frame *frame)
{
    buffer *playbuf = (buffer*)data;
    signed long tagsize;

    switch (stream->error) {
        case MAD_ERROR_BADDATAPTR:
            return MAD_FLOW_CONTINUE;

        case MAD_ERROR_LOSTSYNC:
            tagsize = id3_tag_query(stream->this_frame,
                                    stream->bufend - stream->this_frame);
            if (tagsize > 0) {
                struct id3_tag *tag = NULL;
                // We have an ID3 tag somewhere in the beginning of the stream!
                // We need to skip over it or else we'll get a loud click and a
                // bunch of decoder errors
                tag = get_id3(stream, tagsize, playbuf);
                if (tag) {
                    show_id3(tag);
                    id3_tag_delete(tag);
                }
                return MAD_FLOW_CONTINUE;
            }
            /* fall through */

        default:
            num_decoder_errors++;
#ifdef MAX_DECODER_ERRORS
            if ( num_decoder_errors > MAX_DECODER_ERRORS ) {
                errno = EIO; // read error
                mpg321_error("<<<%s>>>: Too many decode errors!\n", playbuf->filename);
                // Bail
                return MAD_FLOW_BREAK;
            }
            else 
#endif /* MAX_DECODER_ERRORS */                
            {
                if (options.opt & MPG321_REMOTE_PLAY) {
                    mpg321_msg("@E MAD error:%#x: %s\n",
                               stream->error, mad_stream_errorstr(stream)); 
                }
                else {
                    fprintf( stderr, "MAD error: %#x: %s\n", stream->error, mad_stream_errorstr(stream));
                }
            }    
            break;
    }

    return MAD_FLOW_CONTINUE;
}


void usage(char *argv0)
{
    mpg123_boilerplate();
    fprintf(stderr,
        "\nUsage: %s [options] file(s) | URL(s) | -\n\n"
        "Options supported:\n"
        "   --verbose or -v          Increase verbosity\n"
        "   --quiet or -q            Quiet mode (no title or boilerplate)\n"
        "   --gain N or -g N         Set gain (audio volume) to N (0-100)\n"
        "   --skip N or -k N         Skip N frames into the file\n"
        "   --verbose or -v          Be more verbose in playing files\n"
        "   -o dt                    Set output devicetype to dt\n" 
    "                                [esd,alsa(09),arts,sun,oss]\n"
        "   --audiodevice N or -a N  Use N for audio-out\n"
        "   --stdout or -s           Use stdout for audio-out\n"
        "   --au N                   Use au file N for output\n"
        "   --cdr N                  Use wave file N for output\n"
        "   --wav N or -w N          Use wave file N for output\n"
        "   --test or -t             Test only; do no audio output\n"
        "   --list N or -@ N         Use playlist N as list of MP3 files\n"
        "   --random or -Z           Play files randomly until interrupted\n"
        "   --shuffle or -z          Shuffle list of files before playing\n"
        "   --ignore-crc or -i       Tell libmad to ignore CRC errors\n"
        "   -R                       Use remote control interface\n"
        "   --aggressive             Try to get higher priority\n"
        "   --help or --longhelp     Print this help screen\n"
        "   --version or -V          Print version information\n"
        "\n"
        "This version of mpg321 has been configured with " AUDIO_DEFAULT " as its default\n"
        "libao output device.\n" , argv0);
}

/* retsigtype is defined by configure;
   pass handle_signals -1 to initialize it */
RETSIGTYPE handle_signals(int sig) 
{
    static struct timeval last_signal = { 0, 0 };
    struct timeval curr_tv;
    
    gettimeofday(&curr_tv, NULL);
    
    if (sig == SIGINT)
    {
        if (((curr_tv.tv_sec - last_signal.tv_sec) * 1000000 +  
            (curr_tv.tv_usec - last_signal.tv_usec)) < 1000000) /* Allow a short delay to
                                                                  kill mpg321 */
        {
            quit_now = 1;
        }
            
        last_signal.tv_sec = curr_tv.tv_sec;
        last_signal.tv_usec = curr_tv.tv_usec;
        
        stop_playing_file = 1;
   }
    else if (sig == SIGBUS)
    {
        // We may have mmapped a bad file.  Stop playing it
        stop_playing_file = 1;
    }
   else if (sig == -1) /* initialize */
   {
        last_signal.tv_sec = curr_tv.tv_sec;
        last_signal.tv_usec = curr_tv.tv_usec;
   }
}

/*
    Only shows ID3 tags if at least one of
    TITLE, ARTIST, ALBUM, YEAR, COMMENT, GENRE
    is present.
*/
static int show_id3(struct id3_tag const *tag)
{
    unsigned int i;
    int print = 0;
    char emptystring[31];
    char *names[6];
    struct {
        int index;
        char const *id;
        char const *name;
    } const info[] = {
        { 0,    ID3_FRAME_TITLE,  "Title  : "   },
        { 1,    ID3_FRAME_ARTIST, "  Artist: "  },
        { 2,    ID3_FRAME_ALBUM,  "Album  : "   },
        { 3,    ID3_FRAME_YEAR,   "  Year  : "  },
        { 4,    ID3_FRAME_COMMENT,"Comment: "   },
        { 5,    ID3_FRAME_GENRE,  "  Genre : "  }
    };

    memset (emptystring, ' ', 30);
    emptystring[30] = '\0';
    /*  Get ID3 tag if available, 30 chars except 4 for year  */
    for (i=0; i<=5; i++)    {
        names[i] = NULL;
        names[i] = id3_get_tag(tag, info[i].id, (i==3) ? 4 : 30);
    }
    for (i=0; i<=5; i++)    {
        if (names[i])   {
            print = 1;
            break;
        }
    }
    if (!print) {
        return 0;
    }

    if (options.opt & MPG321_REMOTE_PLAY)
    {
        mpg321_msg("@I ID3:");

        for (i=0; i<=5; i++)
        {
            if(!names[i])
            {
                mpg321_msg("%s\t", emptystring);
            }
            
            else
            {
                mpg321_msg("%s\t", names[i]);
                free(names[i]);
            }
        }
        mpg321_msg("\n");
    }
    
    else
    {
        /* Emulate mpg123 original behaviour  */
        for (i=0; i<=5; i++)    {
            fprintf (stderr, "%s", info[i].name);
            if (!names[i])  {
                fprintf (stderr, "%s", emptystring);
            }   else    {
                fprintf (stderr, "%s", names[i]);
                free (names[i]);
            }
            if (i%2) fprintf (stderr, "\n");
        }
    }

    return 1;
}

int main(int argc, char *argv[])
{
    int fd = 0;
    char *currentfile, *currenturl, old_dir[PATH_MAX];
    playlist *pl = NULL;
    struct id3_file *id3struct = NULL;
    struct id3_tag *id3tag = NULL;
    int id3Shown = 0;
    int result = 0;
    size_t map_length = 0;
    int f_mmapped = 0;
    size_t pagesize = getpagesize();

    buffer playbuf;
    
    struct mad_decoder decoder;

    old_dir[0] = '\0';

    playbuf.pl = pl = new_playlist();

    if (!pl)
    {
        fprintf(stderr, "malloc failed at startup!\n");
        exit(1);
    }

#ifdef SOFTWARE_VOLUME
    options.volume = MAD_F_ONE;
#else
    /* Set it to an invalid number.  If user gives -g, we'll set it there */
    options.volume = -1;
#endif
    
    status = MPG321_PLAYING;
    
    /* Get the command line options */
    parse_options(argc, argv, pl);

    /* If there were no files and no playlist specified just print the usage */
    if (!playlist_file && optind == argc)
    {
        usage(argv[0]);
        exit(0);
    }

    if (playlist_file) 
    {
        int err = load_playlist(pl, playlist_file);
        if ( err ) exit(-1);
    }
    
    add_cmdline_files(pl, argv);

    if (shuffle_play)
        shuffle_files(pl);

    ao_initialize();

    check_default_play_device();
    
    if (!(options.opt & MPG321_REMOTE_PLAY))
    {
        handle_signals(-1); /* initialize signal handler */
     
        remote_input_buf[0] = '\0';
    }
    
    if (!(options.opt & MPG321_QUIET_PLAY)) 
        mpg123_boilerplate();
    
    if (options.opt & MPG321_REMOTE_PLAY)
    {
        /* Make sure stdin/stdout are nonblocking! */
        fcntl( 0, F_SETFL, O_NONBLOCK );
        fcntl( 1, F_SETFL, O_NONBLOCK );
        
        mpg321_msg("@R MPG123\n");
    }
    
    /* Play the mpeg files or zip it! */
    while((currenturl = currentfile = get_next_file(pl, &playbuf)))
    {
        if (quit_now) 
            break;
        
        signal(SIGINT, SIG_DFL);

#define FILE_URL_PREFIX "file://"        
        if ( 0 == strncmp(currentfile, FILE_URL_PREFIX, strlen(FILE_URL_PREFIX) ) ) {
            currentfile += strlen(FILE_URL_PREFIX);
        }
        
        playbuf.buf = NULL;
        playbuf.fd = -1;
        playbuf.length = 0;
        playbuf.input_length = 0;
#if 0
        playbuf.content_length = 0;
#endif        
        playbuf.done = 0;
        playbuf.num_frames = 0;
        playbuf.max_frames = -1;
        strncpy(playbuf.filename,currentfile, PATH_MAX);
        playbuf.filename[PATH_MAX-1] = '\0';
        id3Shown = 0;
        num_decoder_errors = 0;
        
        if (status == MPG321_PLAYING) 
            file_change = 1;

        mad_timer_reset(&playbuf.duration);
        
        mad_timer_reset(&current_time);

        if ((!(options.opt & MPG321_QUIET_PLAY) || (options.opt & MPG321_REMOTE_PLAY)) && file_change)
        {
            id3struct = id3_file_open (currentfile, ID3_FILE_MODE_READONLY);

            if (id3struct)
            {
                id3tag = id3_file_tag(id3struct);
            
                if (id3tag)
                {
                    id3Shown = show_id3(id3tag);
                }
                else
                {
                    if ( ! (options.opt & MPG321_REMOTE_PLAY) ) {
                        fprintf(stderr, "Allocation error");
                        exit(1);
                    }
                }

                id3_file_close (id3struct);
            }
        }

        /* Create the MPEG stream */
        /* Check if source is on the network */
        if( (fd = raw_open(currentfile)) != 0 || 
            (fd = http_open(currentfile, &playbuf)) != 0 ||
            (fd = ftp_open(currentfile)) != 0 
#ifndef HAVE_MMAP
            || (fd = open(currentfile, O_RDONLY)) != 0 
#endif
            )
        {
            playbuf.fd = fd;
            playbuf.buf = malloc(BUF_SIZE);
            if ( 0 == playbuf.length ) playbuf.length = BUF_SIZE;

            mad_decoder_init(&decoder, &playbuf, read_from_fd, read_header, decode_filter,
                            output, decode_error, /* message */ 0);
        }

        /* Check if we are to use stdin for input */
        else if(strcmp(currentfile, "-") == 0)
        {
            playbuf.fd = fileno(stdin);
            playbuf.buf = malloc(BUF_SIZE);
            playbuf.length = BUF_SIZE;

            mad_decoder_init(&decoder, &playbuf, read_from_fd, read_header, decode_filter,
                            output, decode_error, /* message */ 0);
        }
            
        /* currentfile is a local file (presumably.) mmap() it */
        else if (NULL == strstr(currentfile, "http:"))
        {
            struct stat stat;
            
            if((fd = open(currentfile, O_RDONLY)) == -1)
            {
                mpg321_error("open failed on <<<%s>>>", currenturl);
                goto cont;
            }

            if(fstat(fd, &stat) == -1)
            {
                mpg321_error("fstat failed on <<<%s>>>", currenturl);
                goto cont;
            }
            
            if (!S_ISREG(stat.st_mode))
            {
                mpg321_error("<<<%s>>>: not a regular file", currentfile);
                goto cont;
            }
            
            calc_length(currentfile, &playbuf);

            if (((long)options.maxframes != (long)-1) &&
                ((long)options.maxframes <= (long)playbuf.num_frames)) 
            { 
                playbuf.max_frames = options.maxframes;
            }
            
            playbuf.frames = malloc((playbuf.num_frames + 1) * sizeof(void*));
            playbuf.times = malloc((playbuf.num_frames + 1) * sizeof(mad_timer_t));
            if ( ( NULL == playbuf.frames ) || ( NULL == playbuf.times ) )
            {
                mpg321_error("%s: num_frames too big!", currenturl);
                goto cont;
            }                

            if ( playbuf.length == 0 ) 
            {
                // Empty file, send "stopped" & "closed"
                errno = EINVAL; // This tells owner what happened
                mpg321_error("<<<%s>>>: empty file", currenturl);
                goto cont;
            }
                
            map_length = (( playbuf.length + (pagesize - 1) ) / pagesize ) * pagesize;
            if((playbuf.buf = mmap(0, map_length, PROT_READ, MAP_SHARED, fd, 0))
               == MAP_FAILED)
            {
                mpg321_error("mmap failed on <<<%s>>>", currenturl);
                goto cont;
            }
            
            f_mmapped = 1;
            
            playbuf.frames[0] = playbuf.buf;
            
            mad_decoder_init(&decoder, &playbuf, read_from_mmap, read_header, decode_filter,
                            output, decode_error, /* message */ 0);
            goto ok;
            
          cont:
            mpg321_msg("@P 0\n");
            close(fd);
            mpg321_msg("@P -1\n");
            clear_remote_file(pl);
            continue;
            
          ok:
            ; // Empty statement to quiet the compiler
        }
        else {
            // We couldn't open anything!  Stop
            clear_remote_file(pl);
            continue;
        }

        if ( options.opt & MPG321_IGNORE_CRC )
        {
            mad_decoder_options(&decoder, MAD_OPTION_IGNORECRC);
        }
        
        if ( options.opt & MPG321_REMOTE_PLAY && ! id3Shown )
        {
            if ( 0 == strncmp( currenturl, "http://", strlen("http://")) )
            {
                mpg321_msg("@I %s\n", currentfile);
            }
            else {
                char * basec = strdup(currentfile);
                char * basen = basename(basec);
                        
                char * dot = strrchr(basen, '.');
                        
                if (dot)
                    *dot = '\0';
                        
                mpg321_msg("@I %s\n", basen);

                free(basec);
            }
        }
    
        if(!(options.opt & MPG321_QUIET_PLAY))/*zip it!!!*/
        {
            /* Because dirname might modify the argument */
            char * dirc = strdup(currentfile);
            char * basec = strdup(currentfile);

            char * basen = basename(basec);
            char * dirn = dirname(dirc);
            
            /* make sure that the file has a pathname; otherwise don't print out
               a Directory: listing */
            if(strchr(currentfile, '/') && strncmp(old_dir, dirn, PATH_MAX) != 0)
            {
                /* Print information about the file */
                fprintf(stderr, "\n");
                fprintf(stderr,"Directory: %s/\n", dirn);
                
                strncpy(old_dir, dirn, PATH_MAX);
                old_dir[PATH_MAX-1] = '\0';
            }
            
            /* print a newline between different songs only, not after
               Directory: listing */
            else
            {
                fprintf(stderr, "\n");
            }

            fprintf(stderr,"Playing MPEG stream from %s ...\n", basen);
    
            free(dirc);
            free(basec);
        }    

        signal(SIGINT, handle_signals);
//        signal(SIGBUS, handle_signals);

        /* This makes SURE we reopen the device on each new song,
           which should alleviate (a little) any problems with someone
           changing the sample rate out from under us
         */
        if(playdevice) {
            ao_close(playdevice);
            playdevice = NULL;
        }
        
        /* Every time the user gets us to rewind, we exit decoding,
           reinitialize it, and re-start it */
        while (1)
        {
            result = mad_decoder_run(&decoder, MAD_DECODER_MODE_SYNC);
            
            /* if we're rewinding on an mmap()ed stream */
            if(( ( status == MPG321_SEEKING && 0 != playbuf.seek_by_time_byte_offset ) || 
                 status == MPG321_REWINDING ) && playbuf.fd == -1) 
            {
                mad_decoder_init(&decoder, &playbuf, read_from_mmap, read_header, decode_filter,
                    output, decode_error, /* message */ 0);
            }    
            else
                break;
        } 

        mpg321_printf("Decoding of %s finished.\n",basename(currentfile));
        
        if (!(options.opt & MPG321_QUIET_PLAY))
        {
            char time_formatted[11];
            mad_timer_string(current_time, time_formatted, "%.1u:%.2u", MAD_UNITS_MINUTES,
                             MAD_UNITS_SECONDS, 0);
            fprintf(stderr, "\n[%s] Decoding of %s finished.\n",time_formatted, basename(currentfile));
        }
        
        if (options.opt & MPG321_REMOTE_PLAY &&
            (status == MPG321_STOPPED || result != 0) )
        {
            clear_remote_file(pl);
        }

        mad_decoder_finish(&decoder);

        if (quit_now)
            break;

        if (playbuf.frames)
            free(playbuf.frames);

        if (playbuf.times)
            free(playbuf.times);
            
        if ( f_mmapped )
        {
            int s = munmap(playbuf.buf, map_length);
            if ( s < 0 ) 
            {
                mpg321_error("munmap failed...restarting");
                exit(-1);  
            }
            
            s = close(fd);
            if ( s < 0 )
            {
                mpg321_error("close failed...restarting");
                exit(-1);
            }
            
            fd = 0;
            if ( status != MPG321_SEEKING && options.opt & MPG321_REMOTE_PLAY ) { 
                mpg321_printf("%s","Closing mmap\n");
                mpg321_msg("@P -1\n");
            }
        }

        else
        {
            free(playbuf.buf);
            if (playbuf.fd != fileno(stdin)) {
                close(playbuf.fd);
                fd = 0;
                if (options.opt & MPG321_REMOTE_PLAY ) { 
                    mpg321_msg("@P -1\n");
                }
            }
        }
    }

    if(playdevice)
        ao_close(playdevice);

    ao_shutdown();

#if defined(RAW_SUPPORT) || defined(HTTP_SUPPORT) || defined(FTP_SUPPORT) 
    if(fd) {
        close(fd);
        if (options.opt & MPG321_REMOTE_PLAY ) { 
            mpg321_msg("@P -1\n");
        }
    }
#endif

    mpg321_msg("@Q\n");
    
    return(0);
}


/* Convenience for retrieving already formatted id3 data
 * what parameter is one of
 *  ID3_FRAME_TITLE
 *  ID3_FRAME_ARTIST
 *  ID3_FRAME_ALBUM
 *  ID3_FRAME_YEAR
 *  ID3_FRAME_COMMENT
 *  ID3_FRAME_GENRE
 * It allocates a new string. Free it later.
 * NULL if no tag or error.
 */
char *id3_get_tag (struct id3_tag const *tag, char const *what, unsigned int maxlen)
{
    struct id3_frame const *frame = NULL;
    union id3_field const *field = NULL;
    int nstrings;
    int avail;
    int j;
    int tocopy;
    int len;
    char printable [1024];
    char *retval = NULL;
    id3_ucs4_t const *ucs4 = NULL;
    id3_latin1_t *latin1 = NULL;

    memset (printable, '\0', 1024);
    avail = 1024;
    if (strcmp (what, ID3_FRAME_COMMENT) == 0)
    {
        /*There may be sth wrong. I did not fully understand how to use
            libid3tag for retrieving comments  */
        j=0;
        frame = id3_tag_findframe(tag, ID3_FRAME_COMMENT, j++);
        if (!frame) return (NULL);
        ucs4 = id3_field_getfullstring (&frame->fields[3]);
        if (!ucs4) return (NULL);
        latin1 = id3_ucs4_latin1duplicate (ucs4);
        if (!latin1 || strlen(latin1) == 0) return (NULL);
        len = strlen(latin1);
        if (avail > len)
            tocopy = len;
        else
            tocopy = 0;
        if (!tocopy) return (NULL);
        avail-=tocopy;
        strncat (printable, latin1, tocopy);
        free (latin1);
    }
    
    else
    {
        frame = id3_tag_findframe (tag, what, 0);
        if (!frame) return (NULL);
        field = &frame->fields[1];
        nstrings = id3_field_getnstrings(field);
        for (j=0; j<nstrings; ++j)
        {
            ucs4 = id3_field_getstrings(field, j);
            if (!ucs4) return (NULL);
            if (strcmp (what, ID3_FRAME_GENRE) == 0)
                ucs4 = id3_genre_name(ucs4);
            latin1 = id3_ucs4_latin1duplicate(ucs4);
            if (!latin1) break;
            len = strlen(latin1);
            if (avail > len)
                tocopy = len;
            else
                tocopy = 0;
            if (!tocopy) break;
            avail-=tocopy;
            strncat (printable, latin1, tocopy);
            free (latin1);
        }
    }
    retval = malloc (maxlen + 1);
    if (!retval) return (NULL);

    strncpy (retval, printable, maxlen);
    retval[maxlen] = '\0';

    len = strlen(printable);
    if (maxlen > (unsigned int)len)
    {
        memset (retval + len, ' ', maxlen - len);
    }

    return (retval);
}
